unit NTSetOwnership;

interface

uses
  Windows, Messages, SysUtils, Classes, Dialogs;

const
  SECURITY_DESCRIPTOR_REVISION = 1;
  SE_TAKE_OWNERSHIP       = 'SeTakeOwnershipPrivilege';
  SE_RESTORE              = 'SeRestorePrivilege';
  SE_BACKUP               = 'SeBackupPrivilege';
  SE_CHANGE_NOTIFY        = 'SeChangeNotifyPrivilege';

type
  TAquiredPrivileges    = (TakeOwnership, RestoreFiles, BackupFiles, BypassTraverseChecking);
  TAquiredPrivilegesSet = Set of TAquiredPrivileges;
  TNTSetOwnership = class(TComponent)
  private
    { Private declarations }
    F_Activate               : Boolean;
    F_Domain                 : String;
    F_UserName               : String;
    F_GroupName              : String;
    F_Path                   : String;
    F_Subdirectories         : Boolean;
    F_LastError              : Integer;
    F_LastErrorText          : String;
    F_AquiredPrivileges      : TAquiredPrivilegesSet;
    UOG                      : String;
    BatchError               : Boolean;
    Function    GetAccountSid(ServerName, AccountName : PChar; Var SID : PSID) : Boolean;
    Function    SetPrivilege(HToken : THandle; Privilege : PChar; EnablePrivilege:Boolean): Boolean;
    Procedure   SetTreeOwner(SDir    : String;
                             SecInfo : SECURITY_INFORMATION;
                             PFileSD : PSECURITY_DESCRIPTOR);
  protected
    { Protected declarations }
    Procedure   F_Set_GroupName(Value : String);
    Procedure   F_Set_UserName(Value : String);
    Procedure   F_Set_AquiredPrivileges(Value : TAquiredPrivilegesSet);
    Procedure   F_Set_Activate(Value : Boolean);
    Procedure   AdjustPriviledges;
    Function    GetLastErr : Integer;
  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;
    destructor  Destroy; override;
    Function    GetSID(DomainName, AccountName:String; Var SID : PSID) : Boolean;
    Function    ChangeFileOwner(FileName:String; OwnerSID : PSID):Boolean;
    Function    Execute : Boolean;
  published
    { Published declarations }
    Property Activate          : Boolean               Read F_Activate          Write F_Set_Activate;
    Property Domain            : String                Read F_Domain            Write F_Domain;
    Property GroupName         : String                Read F_GroupName         Write F_Set_GroupName;
    Property LastError         : Integer               Read F_LastError         Write F_LastError;
    Property LastErrorText     : String                Read F_LastErrorText     Write F_LastErrorText;
    Property Path              : String                Read F_Path              Write F_Path;
    Property Subdirectories    : Boolean               Read F_Subdirectories    Write F_Subdirectories;
    Property UserName          : String                Read F_UserName          Write F_Set_UserName;
    Property AquiredPrivileges : TAquiredPrivilegesSet Read F_AquiredPrivileges Write F_Set_AquiredPrivileges;
  end;

function NetGetDCName (serverName, domainName : PWideChar; var buffer : Pointer) : Integer; stdcall;

procedure Register;


implementation
function NetGetDCName;                  external 'NETAPI32.DLL';

Function TNTSetOwnership.GetLastErr : Integer;
Begin
  Result          := GetLastError;
  F_LastError     := Result;
  F_LastErrorText := SysErrorMessage(F_LastError);
End;

Function TNTSetOwnership.GetAccountSid(ServerName, AccountName : PChar; Var SID : PSID) : Boolean;
Var
  SidSz : DWORD;
  RDSz  : DWORD;
  RD    : PChar;
  Use   : SID_NAME_USE;
Begin
  Result := False;
  RD     := Nil;
  SidSz  := 256;
  RDSz   := 32;
  Try
    SID  := PSID(HeapAlloc(GetProcessHeap, 0, SidSz));
    If SID = Nil Then Exit;
    RD   := PChar(HeapAlloc(GetProcessHeap, 0, RDSz));
    If RD = Nil Then Exit;
    while(Not LookupAccountName(ServerName,
                                AccountName,
                                SID,
                                SidSz,
                                RD,
                                RDSz,
                                Use)) Do
    Begin
      if (GetLastErr = ERROR_INSUFFICIENT_BUFFER) Then
         Begin
           SID  := PSID(HeapReAlloc(GetProcessHeap, 0, SID, SidSz));
           If SID = Nil Then Exit;
           RD   := PChar(HeapReAlloc(GetProcessHeap, 0, RD, RDSz));
           If RD = Nil Then Exit;
         End
       Else
         Begin
           Exit;
         End;
    End;
    Result := True;
  Finally
    if (RD <> Nil) Then HeapFree(GetProcessHeap, 0, RD);
    if (Not Result) And (SID <> Nil) Then HeapFree(GetProcessHeap, 0, SID);
  End;
End;

Function TNTSetOwnership.GetSID(DomainName, AccountName:String; Var SID : PSID) : Boolean;
Var
 PDOMAIN_CONTROLLER_W : Pointer;
 DOMAIN_CONTROLLER_W  : Array[0..1000] of WideChar;
 DOMAIN_W             : Array[0..1000] of WideChar;
 DOMAIN_CONTROLLER_A  : String;
Begin
  if DomainName='' Then
     Begin
       Result := GetAccountSid(Nil, PChar(AccountName), SID);
     End
  Else
     Begin
      MultiByteToWideChar(CP_ACP,0,PChar(Domain),-1,DOMAIN_W,1000);
      PDOMAIN_CONTROLLER_W := @DOMAIN_CONTROLLER_W;
      if (NetGetDCName(Nil,DOMAIN_W,PDOMAIN_CONTROLLER_W) <> ERROR_SUCCESS) Then
         Begin
           Result := GetAccountSid(PChar(DomainName), PChar(AccountName), SID);
         End
      Else
         Begin
          DOMAIN_CONTROLLER_A := WideCharToString(PDOMAIN_CONTROLLER_W);
          Result := GetAccountSid(PChar(DOMAIN_CONTROLLER_A), PChar(AccountName), SID);
         End;
     End;
End;

Function TNTSetOwnership.SetPrivilege(HToken : THandle; Privilege : PChar; EnablePrivilege:Boolean): Boolean;
Var
  TP         : TTOKENPRIVILEGES;
  LUID       : TLargeInteger;
  TPPrevious : TTOKENPRIVILEGES;
  PreviousSz : DWORD;
  DummySz    : DWORD;
  CloseToken : Boolean;
Begin
  Result     := False;
  CloseToken := False;
  PreviousSz := SizeOf(TTOKENPRIVILEGES);

  //////////////////////////////////////////////////////////////////////////////
  if (hToken=0) Then
     Begin
       if (Not OpenProcessToken(GetCurrentProcess,
				TOKEN_ADJUST_PRIVILEGES Or TOKEN_QUERY,
				HToken)) Then
          Begin
            //////////////////////////////////////////// OpenProcessToken failed
            Exit;
          End;
       CloseToken:=True;
     End;

  //////////////////////////////////////////////////////////////////////////////
  if (Not LookupPrivilegeValue(Nil, Privilege, LUID))  Then
     Begin
       ///////////////////////////////////////////// LookupPrivilegeValue failed
       Exit;
     End;

  //////////////////////////////////////////////////////////////////////////////
  TP.PrivilegeCount           := 1;
  TP.Privileges[0].Luid       := LUID;
  TP.Privileges[0].Attributes := 0;
  AdjustTokenPrivileges(HToken,
                        False,
                        TP,
                        SizeOf(TTOKENPRIVILEGES),
                        TPPrevious,
                        PreviousSz);
  if (GetLastErr <> ERROR_SUCCESS) Then
     Begin
       //////////////////////////////////////////// AdjustTokenPrivileges failed
       Exit;
     End;

  //////////////////////////////////////////////////////////////////////////////
  TPPrevious.PrivilegeCount       := 1;
  TPPrevious.Privileges[0].Luid   := LUID;
  DummySz := 0;
  
  if(EnablePrivilege)  Then
    Begin
      TPPrevious.Privileges[0].Attributes := (TPPrevious.Privileges[0].Attributes Or SE_PRIVILEGE_ENABLED);
    End
  Else
    Begin
      TPPrevious.Privileges[0].Attributes := (TPPrevious.Privileges[0].Attributes) XOR (SE_PRIVILEGE_ENABLED AND tpPrevious.Privileges[0].Attributes);
    End;
  AdjustTokenPrivileges(HToken,
                        False,
                        TPPrevious,
                        PreviousSz,
                        Nil,
                        DummySz);

 if (GetLastErr <> ERROR_SUCCESS) Then
    Begin
      //////////////////////////////////////////// AdjustTokenPrivileges failed
      Exit;
    End;
 if (CloseToken) Then CloseHandle(HToken);
 Result := True;
End;

Procedure TNTSetOwnership.SetTreeOwner(SDir    : String;
                                       SecInfo : SECURITY_INFORMATION;
                                       PFileSD : PSECURITY_DESCRIPTOR);
var
  DIR    : TSearchRec;
  DirPath: String;
  Result : Integer;
  SDirTmp: String;
Begin
  if Length(SDir) < 4 Then Exit;
  if BatchError Then Exit;
  DirPath:=SDir+'\*.*';
  Result:=FindFirst(DirPath,faAnyFile,DIR);
  while (Result=0) do
    Begin
     if (DIR.Attr And faDirectory) > 0 Then
         Begin
           if ((DIR.Name <> '.') And (DIR.Name <> '..')) Then
              Begin
                SDirTmp:=SDir+'\'+DIR.Name;
                SetTreeOwner(SDirTmp, SecInfo, PFileSD);
                if (Not SetFileSecurity(PChar(SDirTmp),SecInfo, PFileSD)) Then
                   Begin
                     GetLastErr;
                     BatchError := True;
                     FindClose(DIR);
                     Exit;
                   End;
                ////////////////////////////////////////////////////////////////
              End;
         End
      else
         Begin
           DirPath:=SDir+'\'+Dir.Name;
           if (Not SetFileSecurity(PChar(DirPath),SecInfo, PFileSD)) Then
              Begin
                GetLastErr;
                BatchError := True;
                FindClose(DIR);
                Exit;
              End;
         End;
      Result:=FindNext(DIR);
    End;
  FindClose(DIR);
End;

Function TNTSetOwnership.ChangeFileOwner(FileName:String; OwnerSID : PSID):Boolean;
Var
 TFileSD : Array[0..99] of Byte;
 PFileSD : PSECURITY_DESCRIPTOR;
Begin
 Result := False;
 PFileSD := @TFileSD;
 FillChar(TFileSD,SizeOf(TFileSD),0);
 if (NOT InitializeSecurityDescriptor(PFileSD, SECURITY_DESCRIPTOR_REVISION)) Then Exit;
 if (NOT SetSecurityDescriptorOwner(PFileSD, OwnerSID, False)) Then Exit;
 if (Not IsValidSecurityDescriptor(PFileSD)) Then Exit;
 If DirectoryExists(FileName) Then
    Begin
      if F_Subdirectories Then
         Begin
           BatchError := False;
           SetTreeOwner(FileName,SECURITY_INFORMATION(OWNER_SECURITY_INFORMATION), PFileSD);
           if BatchError Then Exit;
         End
      Else
         Begin
           if (Not SetFileSecurity(PChar(FileName),SECURITY_INFORMATION(OWNER_SECURITY_INFORMATION), PFileSD)) Then Exit;
         End;
    End
 Else
 If FileExists(FileName) Then
    Begin
      if (Not SetFileSecurity(PChar(FileName),SECURITY_INFORMATION(OWNER_SECURITY_INFORMATION), PFileSD)) Then Exit;
    End;
 Result := True;
End;

Procedure TNTSetOwnership.AdjustPriviledges;
Var
 F_TakeOwnership          : Boolean;
 F_RestoreFiles           : Boolean;
 F_BackupFiles            : Boolean;
 F_BypassTraverseChecking : Boolean;
Begin
 F_AquiredPrivileges := [];
 if (Not SetPrivilege(0,SE_TAKE_OWNERSHIP,True)) Then
    Begin
     F_TakeOwnership := False;
    End
  Else
    Begin
     F_TakeOwnership := True;
    End;
 if (Not SetPrivilege(0,SE_RESTORE,True)) Then
    Begin
     F_RestoreFiles := False;
    End
  Else
    Begin
     F_RestoreFiles := True;
    End;
  if (Not SetPrivilege(0,SE_BACKUP,True)) Then
    Begin
     F_BackupFiles := False;
    End
  Else
    Begin
     F_BackupFiles := True;
    End;
  if (Not SetPrivilege(0,SE_CHANGE_NOTIFY,True)) Then
    Begin
     F_BypassTraverseChecking := False;
    End
  Else
    Begin
     F_BypassTraverseChecking := True;
    End;
  If F_TakeOwnership          Then F_AquiredPrivileges := F_AquiredPrivileges+[TakeOwnership];
  If F_RestoreFiles           Then F_AquiredPrivileges := F_AquiredPrivileges+[RestoreFiles];
  If F_BackupFiles            Then F_AquiredPrivileges := F_AquiredPrivileges+[BackupFiles];
  If F_BypassTraverseChecking Then F_AquiredPrivileges := F_AquiredPrivileges+[BypassTraverseChecking];
End;

Function TNTSetOwnership.Execute : Boolean;
Var
 SID  : PSid;
Begin
  //////////////////////////////////////////////////////////////////////////////
  Result := False;
  SID    := Nil;
  UOG    := F_UserName;
  If UOG  = '' Then UOG := F_GroupName;
  If UOG  = '' Then Exit;
  if Path = '' Then Exit;
  if (Not GetSid(F_Domain,UOG,SID)) Then Exit;
  Result := ChangeFileOwner(Path, Sid);
  GetLastErr;
  if Sid <> Nil Then HeapFree(GetProcessHeap, 0, Sid);
End;

constructor TNTSetOwnership.Create(AOwner : TComponent);
Begin
  Inherited Create(AOwner);
  F_Activate          := False;
  F_AquiredPrivileges := [];
  F_Domain            := '';
  F_UserName          := '';
  F_GroupName         := '';
  F_Path              := '';
  F_Subdirectories    := False;
  F_LastError         := 0;
  F_LastErrorText     := '';
  AdjustPriviledges;
End;

destructor TNTSetOwnership.Destroy;
Begin
 Inherited Destroy;
End;

Procedure TNTSetOwnership.F_Set_GroupName(Value : String);
Begin
 F_GroupName := Value;
 if csDesigning in ComponentState Then Exit;
 if F_GroupName <> '' Then F_UserName := '';
End;

Procedure TNTSetOwnership.F_Set_UserName(Value : String);
Begin
 F_UserName := Value;
 if csDesigning in ComponentState Then Exit;
 if F_UserName <> '' Then F_GroupName := '';
End;

Procedure TNTSetOwnership.F_Set_AquiredPrivileges(Value : TAquiredPrivilegesSet);
Begin
 //********************************************************** REad Only
End;

Procedure TNTSetOwnership.F_Set_Activate(Value : Boolean);
Begin
  F_Activate := Value;
  if csLoading in ComponentState Then Exit;
  if F_Activate Then F_Activate := Execute;
  if csDesigning in ComponentState Then
     Begin
       if F_Activate Then
          ShowMessage('Successfuly set owner of '+F_Path+' to '+UOG)
       Else
          ShowMessage('Error setting owner of '+F_Path+' to '+UOG);
       F_Activate := False;
     End;
End;

procedure Register;
begin
  RegisterComponents('System', [TNTSetOwnership]);
end;

end.
